import itertools

import torch

import pytest
import torchlayers as tl


def single_module(regularization: str, name: str):
    return getattr(tl, regularization)(tl.Linear(10), weight_decay=1000000, name=name)


def multiple_modules(regularization: str, name: str):
    return getattr(tl, regularization)(
        torch.nn.Sequential(
            tl.Linear(40), tl.ReLU(), tl.Linear(20), tl.ReLU(), tl.Linear(10)
        ),
        weight_decay=1000000,
        name=name,
    )


def generate_inputs():
    modules = (single_module, multiple_modules)
    names = ("bias", "weight", None)
    regularizations = ("L1", "L2")
    steps = (1, 3)
    for module, regularization, name, steps in itertools.product(
        modules, regularizations, names, steps
    ):
        yield module(regularization, name), name, steps


def check_gradient_correctness(module, name):
    for param_name, param in module.named_parameters():
        if name is not None:
            if name in param_name:
                assert torch.abs(param.grad).sum() > 10000.0
            else:
                assert torch.abs(param.grad).sum() < 10000.0
        else:
            assert torch.abs(param.grad).sum() > 10000.0


@pytest.mark.parametrize("module,name,steps", list(generate_inputs()))
def test_weight_decay(module, name, steps: bool):
    tl.build(module, torch.randn(1, 20))
    for _ in range(steps):
        output = module(torch.randn(5, 20))
        output.sum().backward()
    check_gradient_correctness(module, name)
    module.zero_grad()
    output = module(torch.rand(5, 20))
    output.sum().backward()
    check_gradient_correctness(module, name)



import torch

import pytest
import torchlayers as tl


@pytest.mark.parametrize("klass", ("Swish", "HardSwish", "HardSigmoid"))
def test_object(klass):
    getattr(tl, klass)()(torch.randn(4, 5, 6))


@pytest.mark.parametrize("function", ("swish", "hard_swish", "hard_sigmoid"))
def test_functional(function):
    getattr(tl, function)(torch.randn(4, 5, 6))



import typing

import torch

from . import convolution


class ConvPixelShuffle(torch.nn.Module):
    """Two dimensional convolution with ICNR initialization followed by PixelShuffle.

    Increases `height` and `width` of `input` tensor by scale, acts like
    learnable upsampling. Due to `ICNR weight initialization <https://arxiv.org/abs/1707.02937>`__
    of `convolution` it has similar starting point to nearest neighbour upsampling.

    `kernel_size` got a default value of `3`, `upscale_factor` got a default
    value of `2`.

    Example::

        import torchlayers as tl


        class MiniAutoEncoder(tl.Module):
            def __init__(self, out_channels):
                super().__init__()
                self.conv1 = tl.Conv(64)
                # Twice smaller image by default
                self.pooling = tl.MaxPool()
                # Twice larger (upscale_factor=2) by default
                self.upsample = tl.ConvPixelShuffle(out_channels)

            def forward(self, x):
                x = self.conv1(x)
                pooled = self.pooling(x)
                return self.upsample(pooled)


        out_channels = 3
        network = MiniAutoEncoder(out_channels)
        tl.build(network, torch.randn(1, out_channels, 64, 64))
        assert network(torch.randn(5, out_channels, 64, 64)).shape == [5, out_channels, 64, 64]

    .. note::

        Currently only `4D` input is allowed (`[batch, channels, height, width]`),
        due to `torch.nn.PixelShuffle` not supporting `3D` or `5D` inputs.
        See [this PyTorch PR](https://github.com/pytorch/pytorch/pull/6340/files)
        for example of dimension-agnostic implementation.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input image
    out_channels : int
        Number of channels produced after PixelShuffle
    upscale_factor : int, optional
        Factor to increase spatial resolution by. Default: `2`
    kernel_size : int or tuple, optional
        Size of the convolving kernel. Default: `3`
    stride : int or tuple, optional
        Stride of the convolution. Default: 1
    padding: int or tuple, optional
        Zero-padding added to both sides of the input. Default: 0
    padding_mode: string, optional
        Accepted values `zeros` and `circular` Default: `zeros`
    dilation: int or tuple, optional
        Spacing between kernel elements. Default: 1
    groups: int, optional
        Number of blocked connections from input channels to output channels. Default: 1
    bias: bool, optional
        If ``True``, adds a learnable bias to the output. Default: ``True``
    initializer: typing.Callable[[torch.Tensor,], torch.Tensor], optional
        Initializer for ICNR initialization, can be a function from `torch.nn.init`.
        Receive tensor as argument and returns tensor after initialization.
        Default: `torch.nn.init.kaiming_normal_`

    """

    def __init__(
        self,
        in_channels,
        out_channels,
        upscale_factor: int = 2,
        kernel_size: int = 3,
        stride: int = 1,
        padding="same",
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
        initializer=None,
    ):
        super().__init__()
        self.convolution = convolution.Conv(
            in_channels,
            out_channels * upscale_factor * upscale_factor,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            padding_mode,
        )

        self.upsample = torch.nn.PixelShuffle(upscale_factor)
        if initializer is None:
            self.initializer = torch.nn.init.kaiming_normal_
        else:
            self.initializer = initializer

    def post_build(self):
        """Initialize weights after layer was built."""
        self.icnr_initialization(self.convolution.weight.data)

    def icnr_initialization(self, tensor):
        """ICNR initializer for checkerboard artifact free sub pixel convolution.

        Originally presented in
        `Checkerboard artifact free sub-pixel convolution: A note on sub-pixel convolution, resize convolution and convolution resize <https://arxiv.org/abs/1707.02937>`__
        Initializes convolutional layer prior to `torch.nn.PixelShuffle`.
        Weights are initialized according to `initializer` passed to to `__init__`.

        Parameters
        ----------
        tensor: torch.Tensor
                Tensor to be initialized using ICNR init.

        Returns
        -------
        torch.Tensor
                Tensor initialized using ICNR.

        """

        if self.upsample.upscale_factor == 1:
            return self.initializer(tensor)

        new_shape = [int(tensor.shape[0] / (self.upsample.upscale_factor ** 2))] + list(
            tensor.shape[1:]
        )

        subkernel = self.initializer(torch.zeros(new_shape)).transpose(0, 1)

        kernel = subkernel.reshape(subkernel.shape[0], subkernel.shape[1], -1).repeat(
            1, 1, self.upsample.upscale_factor ** 2
        )

        return kernel.reshape([-1, tensor.shape[0]] + list(tensor.shape[2:])).transpose(
            0, 1
        )

    def forward(self, inputs):
        return self.upsample(self.convolution(inputs))



import typing

import torch

from . import module


class _GlobalPool(torch.nn.Module):
    def forward(self, inputs):
        return self._pooling(inputs).reshape(inputs.shape[0], -1)


class GlobalMaxPool1d(_GlobalPool):
    """Applies a 1D global max pooling over the last dimension.

    Usually used after last `Conv1d` layer to get maximum feature values
    for each timestep.

    Internally operates as `torch.nn.AdaptiveMaxPool1d` with redundant `1` dimensions
    flattened.

    Returns
    -------
    `torch.Tensor`
        `2D` tensor `(batch, features)`

    """

    def __init__(self):
        super().__init__()
        self._pooling = torch.nn.AdaptiveMaxPool1d(1)


class GlobalMaxPool2d(_GlobalPool):
    """Applies a 2D global max pooling over the last dimension(s).

    Usually used after last `Conv2d` layer to get maximum value feature values
    for each channel. Can be used on `3D` or `4D` input (though the latter is more common).

    Internally operates as `torch.nn.AdaptiveMaxPool2d` with redundant `1` dimensions
    flattened.

    Returns
    -------
    `torch.Tensor`
        `2D` tensor `(batch, features)`

    """

    def __init__(self):
        super().__init__()
        self._pooling = torch.nn.AdaptiveMaxPool2d(1)


class GlobalMaxPool3d(_GlobalPool):
    """Applies a 3D global max pooling over the last dimension(s).

    Usually used after last `Conv3d` layer to get maximum value feature values
    for each channel. Can be used on `4D` or `5D` input (though the latter is more common).

    Internally operates as `torch.nn.AdaptiveMaxPool3d` with redundant `1` dimensions
    flattened.

    Returns
    -------
    `torch.Tensor`
        `2D` tensor `(batch, features)`

    """

    def __init__(self):
        super().__init__()
        self._pooling = torch.nn.AdaptiveMaxPool3d(1)


class GlobalAvgPool1d(_GlobalPool):
    """Applies a 1D global average pooling over the last dimension.

    Usually used after last `Conv1d` layer to get mean of features values
    for each timestep.

    Internally operates as `torch.nn.AdaptiveAvgPool1d` with redundant `1` dimensions
    flattened.

    Returns
    -------
    `torch.Tensor`
        `2D` tensor `(batch, features)`

    """

    def __init__(self):
        super().__init__()
        self._pooling = torch.nn.AdaptiveAvgPool1d(1)


class GlobalAvgPool2d(_GlobalPool):
    """Applies a 2D global average pooling over the last dimension(s).

    Usually used after last `Conv2d` layer to get mean value of features values
    for each channel. Can be used on `3D` or `4D` input (though the latter is more common).

    Internally operates as `torch.nn.AdaptiveAvgPool3d` with redundant `1` dimensions
    flattened.

    Returns
    -------
    `torch.Tensor`
        `2D` tensor `(batch, features)`

    """

    def __init__(self):
        super().__init__()
        self._pooling = torch.nn.AdaptiveAvgPool2d(1)


class GlobalAvgPool3d(_GlobalPool):
    """Applies a 3D global average pooling over the last dimension(s).

    Usually used after last `Conv3d` layer to get mean value of features values
    for each channel. Can be used on `4D` or `5D` input (though the latter is more common).

    Internally operates as `torch.nn.AdaptiveAvgPool3d` with redundant `1` dimensions
    flattened.

    Returns
    -------
    `torch.Tensor`
        `2D` tensor `(batch, features)`

    """

    def __init__(self):
        super().__init__()
        self._pooling = torch.nn.AdaptiveAvgPool3d(1)


class GlobalMaxPool(module.InferDimension):
    """Perform `max` pooling operation leaving maximum values from channels.

    Usually used after last convolution layer (`torchlayers.Conv`)
    to get pixels of maximum value from each channel.

    Depending on shape of passed `torch.Tensor` either `1D`, `2D` or `3D` `GlobalMaxPool`
    will be used for `3D`, `4D` and `5D` shape respectively (batch included).

    Internally operates as `torchlayers.pooling.GlobalMaxPoolNd`.

    Returns
    -------
    `torch.Tensor`
        `2D` tensor `(batch, features)`

    """

    def __init__(self):
        super().__init__(
            dispatcher={5: GlobalMaxPool3d, 4: GlobalMaxPool2d, 3: GlobalMaxPool1d}
        )


class GlobalAvgPool(module.InferDimension):
    """Perform `mean` pooling operation leaving average values from channels.

    Usually used after last convolution layer (`torchlayers.Conv`) to get mean
    of pixels from each channel.

    Depending on shape of passed `torch.Tensor` either `1D`, `2D` or `3D`
    pooling will be used for `3D`, `4D` and `5D`
    shape respectively (batch included).

    Internally operates as `torchlayers.pooling.GlobalAvgPoolNd`.

    Returns
    -------
    `torch.Tensor`
        `2D` tensor `(batch, features)`

    """

    def __init__(self):
        super().__init__(
            dispatcher={5: GlobalAvgPool3d, 4: GlobalAvgPool2d, 3: GlobalAvgPool1d}
        )


class MaxPool(module.InferDimension):
    """Perform `max` operation across first `torch.Tensor` dimension.

    Depending on shape of passed `torch.Tensor` either `torch.nn.MaxPool1D`,
    `torch.nn.MaxPool2D` or `torch.nn.MaxPool3D` pooling will be used
    for `3D`, `4D` and `5D` shape respectively (batch included).

    Default value for `kernel_size` (`2`) was added.

    Parameters
    ----------
    kernel_size: int, optional
        The size of the window to take a max over. Default: `2`
    stride: int, optional
        The stride of the window. Default value is :attr:`kernel_size`
    padding: int, optional
        Implicit zero padding to be added on both sides. Default: `0`
    dilation: int
        Parameter controlling the stride of elements in the window. Default: `1`
    return_indices: bool, optional
        If ``True``, will return the max indices along with the outputs.
        Useful for :class:`torch.nn.MaxUnpool` later. Default: `False`
    ceil_mode: bool, optional
        When True, will use `ceil` instead of `floor` to compute the output shape.
        Default: `False`

    Returns
    -------
    `torch.Tensor`
        Same shape as `input` with values pooled.

    """

    def __init__(
        self,
        kernel_size: int = 2,
        stride: int = None,
        padding: int = 0,
        dilation: int = 1,
        return_indices: bool = False,
        ceil_mode: bool = False,
    ):
        super().__init__(
            dispatcher={
                5: torch.nn.MaxPool3d,
                4: torch.nn.MaxPool2d,
                3: torch.nn.MaxPool1d,
            },
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            return_indices=return_indices,
            ceil_mode=ceil_mode,
        )


class AvgPool(module.InferDimension):
    """Perform `avg` operation across first `torch.Tensor` dimension.

    Depending on shape of passed `torch.Tensor` either `torch.nn.AvgPool1D`,
    `torch.nn.AvgPool2D` or `torch.nn.AvgPool3D` pooling will be used
    for `3D`, `4D` and `5D` shape respectively (batch included).

    Default value for `kernel_size` (`2`) was added.

    Parameters
    ----------
    kernel_size: int, optional
        The size of the window. Default: `2`
    stride: int, optional
        The stride of the window. Default value is :attr:`kernel_size`
    padding: int, oprtional
        Implicit zero padding to be added on both sides. Default: `0`
    ceil_mode: bool, opriontal
        When True, will use `ceil` instead of `floor` to compute the output shape.
        Default: `True`
    count_include_pad: bool, optional
        When True, will include the zero-padding in the averaging. Default: `True`

    Returns
    -------
    `torch.Tensor`
        Same shape as `input` with values pooled.

    """

    def __init__(
        self,
        kernel_size: int = 2,
        stride: int = None,
        padding: int = 0,
        ceil_mode: bool = False,
        count_include_pad: bool = True,
    ):
        super().__init__(
            dispatcher={
                5: torch.nn.AvgPool3d,
                4: torch.nn.AvgPool2d,
                3: torch.nn.AvgPool1d,
            },
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            ceil_mode=ceil_mode,
            count_include_pad=count_include_pad,
        )



import collections
import inspect
import io
import types
import typing
import warnings

import torch

from . import (_dev_utils, activations, convolution, normalization, pooling,
               preprocessing, regularization, upsample)
from ._version import __version__
from .module import InferDimension

__all__ = ["summary", "build", "infer", "Lambda", "Reshape", "Concatenate"]


def summary(module, *args, **kwargs):
    """Create summary of PyTorch module.

    Works similarly to `summary` functionality provided by `keras`.

    Example::

        import torchlayers as tl


        class Classifier(tl.Module):
            def __init__(self):
                super().__init__()
                self.conv1 = tl.Conv2d(64, kernel_size=6)
                self.conv2 = tl.Conv2d(128, kernel_size=3)
                self.conv3 = tl.Conv2d(256, kernel_size=3, padding=1)
                self.pooling = tl.GlobalMaxPool()
                self.dense = tl.Linear(10)

            def forward(self, x):
                x = torch.relu(self.conv1(x))
                x = torch.relu(self.conv2(x))
                x = torch.relu(self.conv3(x))
                return self.dense(self.pooling(x))

        clf = Classifier()
        tl.build(clf, torch.randn(1, 3, 32, 32))
        # You have to print the summary explicitly
        print(tl.summary(clf, torch.randn(1, 3, 32, 32)))

    Above would print (please notice explicit usage of `print` statement)::


                Layer (type)       |      Inputs      |     Outputs      | Params (size in MB)  | Buffers (size in MB)
        ===============================================================================================================
        ---------------------------------------------------------------------------------------------------------------
               conv1 (Conv2d)      |  (1, 3, 32, 32)  | (1, 64, 27, 27)  |  6976 (0.027904 MB)  |   0 (0.000000 MB)
        ---------------------------------------------------------------------------------------------------------------
               conv2 (Conv2d)      | (1, 64, 27, 27)  | (1, 128, 25, 25) | 73856 (0.295424 MB)  |   0 (0.000000 MB)
        ---------------------------------------------------------------------------------------------------------------
               conv3 (Conv2d)      | (1, 128, 25, 25) | (1, 256, 25, 25) | 295168 (1.180672 MB) |   0 (0.000000 MB)
        ---------------------------------------------------------------------------------------------------------------
         pooling (GlobalMaxPool2d) | (1, 256, 25, 25) |     (1, 256)     |   0 (0.000000 MB)    |   0 (0.000000 MB)
        ---------------------------------------------------------------------------------------------------------------
               dense (Linear)      |     (1, 256)     |     (1, 10)      |  2570 (0.010280 MB)  |   0 (0.000000 MB)
        ---------------------------------------------------------------------------------------------------------------
        ===============================================================================================================
        Total params: 378570 (1.514280 Mb)
        Total buffers: 0 (0.000000 Mb)


    Custom `Summarizer` object is returned from the function. Users can specify which columns to return
    (if any) using `string` function.
    Calling `summary` on the above model like this::

        print(tl.summary(clf, torch.randn(1, 3, 32, 32)).string(buffers=False, inputs=False))

    Would give the following output to `stdout`:


                Layer (type)       |     Outputs      | Params (size in MB)
        =====================================================================
        ---------------------------------------------------------------------
               conv1 (Conv2d)      | (1, 64, 27, 27)  |  6976 (0.027904 MB)
        ---------------------------------------------------------------------
               conv2 (Conv2d)      | (1, 128, 25, 25) | 73856 (0.295424 MB)
        ---------------------------------------------------------------------
               conv3 (Conv2d)      | (1, 256, 25, 25) | 295168 (1.180672 MB)
        ---------------------------------------------------------------------
         pooling (GlobalMaxPool2d) |     (1, 256)     |   0 (0.000000 MB)
        ---------------------------------------------------------------------
               dense (Linear)      |     (1, 10)      |  2570 (0.010280 MB)
        ---------------------------------------------------------------------
        =====================================================================
        Total params: 378570 (1.514280 Mb)
        Total buffers: 0 (0.000000 Mb)

    `string` method has following flags (if `True` the column is returned, `True` by default)
    specifiable by user:

        - layers
        - inputs
        - outputs
        - params
        - buffers

    Returned `Summarizer` object has following fields which can be accessed and
    read/modified by user:

        - names - `list` containing names of consecutive modules
        - modules - `list` containing names of **classes** of consecutive modules
        - inputs - `list` containing inputs (possibly multiple of them) to layers.
        Each `input` is described by `shape` (if it's a `torch.Tensor` instance)
        or by class name otherwise
        - outputs - `list` containing outputs (possibly multiple of them) of layers.
        Each `output` is described by `shape` (if it's a `torch.Tensor` instance)
        or by class name otherwise
        - params - `list` containing number of parameters for each layer
        - params_sizes - `list` containing sizes of parameters for layers (in megabytes)
        - buffers - `list` containing number of buffer parameters for each layer
        - buffers_sizes - `list` containing sizes of buffers for layers (in megabytes)

    Finally, `Summarizer` can also be called with other inputs so it gathers
    representation based on the last provided input. For example::

        summarizer = tl.summary(module, torch.randn(1, 3, 32, 32))
        with summarizer:
            summarizer(torch.randn(5, 3, 64, 64))

        # Inputs and outputs summary for input of shape (5, 3, 64, 64)
        print(summarizer)

    .. note::
            This function can be called before or after `torchlayers.build`
            but the results may vary (names and types of layers before and after
            inference).

    .. note::
            This function runs input through the network hence the input shapes
            are inferred after this call. Network will not be built into
            PyTorch counterparts though (e.g. `tl.Conv` will not become `torch.nn.Conv2d` or a-like).


    Parameters
    ----------
    module : torch.nn.Module
        Instance of module to create summary for
    *args
        Arguments required by module's `forward`
    **kwargs
        Keyword arguments required by module's `forward`
    """

    class Summarizer:
        def __init__(self, module):
            self.module = module

            self.names = []
            self.modules = []

            self.inputs = []
            self.outputs = []

            self.params = []
            self.params_sizes = []

            self.buffers = []
            self.buffers_sizes = []

            self._handles = []
            self._last_module_state = None

        def __enter__(self):
            self._last_module_state = self.module.training
            self.module.eval()
            for name, submodule in list(self.module.named_modules())[1:]:
                if "." not in name:
                    self.names.append(name)
                    self.modules.append(type(submodule).__name__)
                    self._handles.append(submodule.register_forward_hook(self.hook))
            return self

        def __exit__(self, *_, **__):
            for handle in self._handles:
                handle.remove()
            self.module.train(self._last_module_state)
            return False

        @staticmethod
        def _parse(item):
            if torch.is_tensor(item):
                return tuple(item.shape)

            if isinstance(item, collections.abc.Iterable):
                return Summarizer._unwrap(
                    tuple(Summarizer._parse(element) for element in item)
                )
            else:
                return type(item).__name__

        @staticmethod
        def _unwrap(item):
            if len(item) == 1:
                return item[0]
            return item

        def hook(self, module, inputs, outputs):
            def _elements_count(iterator):
                return sum(element.numel() for element in iterator)

            def _elements_size_in_mb(iterator):
                return (
                    sum(
                        element.numel() * element.element_size() for element in iterator
                    )
                    / 1_000_000
                )

            # Inputs and outputs
            self.inputs.append(Summarizer._parse(inputs))
            self.outputs.append(Summarizer._parse(outputs))

            # Parameters and buffers
            self.params.append(_elements_count(module.parameters()))
            self.params_sizes.append(_elements_size_in_mb(module.parameters()))
            self.buffers.append(_elements_count(module.buffers()))
            self.buffers_sizes.append(_elements_size_in_mb(module.buffers()))

        def __call__(self, *args, **kwargs):
            with torch.no_grad():
                return self.module(*args, **kwargs)
            return self

        def string(
            self,
            layers: bool = True,
            inputs: bool = True,
            outputs: bool = True,
            params: bool = True,
            buffers: bool = True,
        ):
            def create_text():
                def conditional_return(conditions, values):
                    for condition, value in zip(conditions, values):
                        if condition:
                            yield value

                def footer(
                    total_params, total_params_sizes, total_buffers, total_buffers_sizes
                ):
                    return "Total params: {} ({:.6f} MB)\nTotal buffers: {} ({:.6f} MB)".format(
                        total_params,
                        total_params_sizes,
                        total_buffers,
                        total_buffers_sizes,
                    )

                layers_text = ["Layer (type)"]
                inputs_text = ["Inputs"]
                outputs_text = ["Outputs"]
                params_text = ["Params (size in MB)"]
                buffers_text = ["Buffers (size in MB)"]

                total_params, total_params_sizes, total_buffers, total_buffers_sizes = (
                    0,
                    0,
                    0,
                    0,
                )
                for (
                    name,
                    module,
                    inp,
                    out,
                    param,
                    param_size,
                    buffer,
                    buffer_size,
                ) in zip(
                    self.names,
                    self.modules,
                    self.inputs,
                    self.outputs,
                    self.params,
                    self.params_sizes,
                    self.buffers,
                    self.buffers_sizes,
                ):

                    layers_text.append("{} ({})".format(name, module))
                    inputs_text.append("{}".format(inp))
                    outputs_text.append("{}".format(out))
                    params_text.append("{} ({:.6f} MB)".format(param, param_size))
                    buffers_text.append("{} ({:.6f} MB)".format(buffer, buffer_size))
                    total_params += param
                    total_params_sizes += param_size
                    total_buffers += buffer
                    total_buffers_sizes += buffer_size

                return (
                    footer(
                        total_params,
                        total_params_sizes,
                        total_buffers,
                        total_buffers_sizes,
                    ),
                    conditional_return(
                        (layers, inputs, outputs, params, buffers),
                        (
                            layers_text,
                            inputs_text,
                            outputs_text,
                            params_text,
                            buffers_text,
                        ),
                    ),
                )

            def center(columns):
                # + 2 to leave some space for representation on both sides
                for column in columns:
                    longest = max(map(len, column)) + 2
                    yield tuple(map(lambda string: string.center(longest), column))

            def join(columns):
                return tuple("|".join(row) for row in zip(*columns))

            footer, columns = create_text()
            rows = join(center(columns))
            representation = rows[0]
            representation += "\n" + "=" * len(rows[0]) + "\n"
            representation += "-" * len(rows[0]) + "\n"
            for row in rows[1:]:
                representation += row
                representation += "\n" + "-" * len(row) + "\n"

            representation += "=" * len(rows[0]) + "\n"
            representation += footer
            return representation

        def __str__(self):
            return self.string()

    with Summarizer(module) as summarizer:
        summarizer(*args, **kwargs)
        return summarizer


def build(module, *args, **kwargs):
    """Build PyTorch layer or module by providing example input.

    This method should be used **always** after creating module using `torchlayers`
    and shape inference especially.

    Works similarly to `build` functionality provided by `keras`.

    Provided module will be "compiled" to PyTorch primitives to remove any
    overhead.

    `torchlayers` also supports `post_build` function to perform some action after
    shape was inferred (weight initialization example below)::


        import torch
        import torchlayers as tl

        class _MyModuleImpl(torch.nn.Linear):
            def post_build(self):
                # You can do anything here really
                torch.nn.init.eye_(self.weights)

        MyModule = tl.infer(_MyModuleImpl)

    `post_build` should have no arguments other than `self` so all necessary
    data should be saved in `module` beforehand.

    Parameters
    ----------
    module : torch.nn.Module
        Instance of module to build
    *args
        Arguments required by module's `forward`
    **kwargs
        Keyword arguments required by module's `forward`
    """

    def torch_compile(module):
        with io.BytesIO() as buffer:
            torch.save(module, buffer)
            return torch.load(io.BytesIO(buffer.getvalue()))

    def run_post(module):
        for submodule in module.modules():
            function = getattr(submodule, "post_build", None)
            if function is not None:
                post_build = getattr(submodule, "post_build")
                if not callable(post_build):
                    raise ValueError(
                        "{}'s post_build is required to be a method.".format(submodule)
                    )
                submodule.post_build()

    with torch.no_grad():
        module.eval()
        module(*args, **kwargs)
    module.train()
    module = torch_compile(module)
    run_post(module)
    return module


def infer(module_class, index: str = 1):
    """Allows custom user modules to infer input shape.

    Input shape should be the first argument after `self`.

    Usually used as class decorator, e.g.::

        import torch
        import torchlayers as tl

        class _StrangeLinearImpl(torch.nn.Linear):
            def __init__(self, in_features, out_features, bias: bool = True):
                super().__init__(in_features, out_features, bias)
                self.params = torch.nn.Parameter(torch.randn(out_features))

            def forward(self, inputs):
                super().forward(inputs) + self.params

        # Now you can use shape inference of in_features
        StrangeLinear = tl.infer(_StrangeLinearImpl)

        # in_features can be inferred
        layer = StrangeLinear(out_features=64)


    Parameters
    ----------
    module_class: torch.nn.Module
        Class of module to be updated with shape inference capabilities.

    index: int, optional
        Index into `tensor.shape` input which should be inferred, e.g. tensor.shape[1].
        Default: `1` (`0` being batch dimension)

    """

    init_arguments = [
        str(argument)
        for argument in inspect.signature(module_class.__init__).parameters.values()
    ]

    # Other argument than self
    if len(init_arguments) > 1:
        name = module_class.__name__
        infered_module = type(
            name, (torch.nn.Module,), {_dev_utils.infer.MODULE_CLASS: module_class},
        )
        parsed_arguments, uninferable_arguments = _dev_utils.infer.parse_arguments(
            init_arguments, infered_module
        )

        setattr(
            infered_module, "__init__", _dev_utils.infer.create_init(parsed_arguments),
        )

        setattr(
            infered_module,
            "forward",
            _dev_utils.infer.create_forward(
                _dev_utils.infer.MODULE,
                _dev_utils.infer.MODULE_CLASS,
                parsed_arguments,
                index,
            ),
        )
        setattr(
            infered_module,
            "__repr__",
            _dev_utils.infer.create_repr(
                _dev_utils.infer.MODULE, **uninferable_arguments
            ),
        )
        setattr(
            infered_module,
            "__getattr__",
            _dev_utils.infer.create_getattr(_dev_utils.infer.MODULE),
        )

        setattr(
            infered_module,
            "__reduce__",
            _dev_utils.infer.create_reduce(_dev_utils.infer.MODULE, parsed_arguments),
        )

        return infered_module

    return module_class


class Lambda(torch.nn.Module):
    """Use any function as `torch.nn.Module`

    Simple proxy which allows you to use your own custom in
    `torch.nn.Sequential` and other requiring `torch.nn.Module` as input::

        import torch
        import torchlayers as tl

        model = torch.nn.Sequential(tl.Lambda(lambda tensor: tensor ** 2))
        model(torch.randn(64 , 20))

    Parameters
    ----------
    function : Callable
        Any user specified function

    Returns
    -------
    Any
        Anything `function` returns

    """

    def __init__(self, function: typing.Callable):
        super().__init__()
        self.function: typing.Callable = function

    def forward(self, *args, **kwargs) -> typing.Any:
        return self.function(*args, **kwargs)


class Concatenate(torch.nn.Module):
    """Concatenate list of tensors.

    Mainly useful in `torch.nn.Sequential` when previous layer returns multiple
    tensors, e.g.::

        import torch
        import torchlayers as tl

        class Foo(torch.nn.Module):
            # Return same tensor three times
            # You could explicitly return a list or tuple as well
            def forward(tensor):
                return tensor, tensor, tensor


        model = torch.nn.Sequential(Foo(), tl.Concatenate())
        model(torch.randn(64 , 20))

    All tensors must have the same shape (except in the concatenating dimension).

    Parameters
    ----------
    dim : int
        Dimension along which tensors will be concatenated

    Returns
    -------
    torch.Tensor
        Concatenated tensor along specified `dim`.

    """

    def __init__(self, dim: int):
        super().__init__()
        self.dim: int = dim

    def forward(self, inputs):
        return torch.cat(inputs, dim=self.dim)


class Reshape(torch.nn.Module):
    """Reshape tensor excluding `batch` dimension

    Reshapes input `torch.Tensor` features while preserving batch dimension.
    Standard `torch.reshape` values (e.g. `-1`) are supported, e.g.::

        import torch
        import torchlayers as tl

        layer = tl.Reshape(20, -1)
        layer(torch.randn(64, 80)) # shape (64, 20, 4)

    All tensors must have the same shape (except in the concatenating dimension).
    If possible, no copy of `tensor` will be performed.

    Parameters
    ----------
    shapes: *int
        Variable length list of shapes used in view function

    Returns
    -------
    torch.Tensor
        Concatenated tensor

    """

    def __init__(self, *shapes: int):
        super().__init__()
        self.shapes: typing.Tuple[int] = shapes

    def forward(self, tensor):
        return torch.reshape(tensor, (tensor.shape[0], *self.shapes))


###############################################################################
#
#                       MODULE AND SHAPE INFERENCE
#
###############################################################################


modules_map = {
    # PyTorch specific tensor index
    "Linear": (torch.nn, -1),
    "RNN": (torch.nn, 2),
    "LSTM": (torch.nn, 2),
    "GRU": (torch.nn, 2),
    "MultiheadAttention": (torch.nn, 2),
    "Transformer": (torch.nn, 2),
    "TransformerEncoderLayer": (torch.nn, 2),
    "TransformerDecoderLayer": (torch.nn, 2),
    # PyTorch default (1) tensor index
    "RNNCell": torch.nn,
    "LSTMCell": torch.nn,
    "GRUCell": torch.nn,
    "Conv1d": torch.nn,
    "Conv2d": torch.nn,
    "Conv3d": torch.nn,
    "ConvTranspose1d": torch.nn,
    "ConvTranspose2d": torch.nn,
    "ConvTranspose3d": torch.nn,
    "BatchNorm1d": torch.nn,
    "BatchNorm2d": torch.nn,
    "BatchNorm3d": torch.nn,
    "SyncBatchNorm": torch.nn,
    "InstanceNorm1d": torch.nn,
    "InstanceNorm2d": torch.nn,
    "InstanceNorm3d": torch.nn,
    # Torchlayers convolution
    "SqueezeExcitation": convolution,
    "Fire": convolution,
    "Conv": convolution,
    "ConvTranspose": convolution,
    "DepthwiseConv": convolution,
    "SeparableConv": convolution,
    "InvertedResidualBottleneck": convolution,
    # Torchlayers normalization
    "BatchNorm": normalization,
    "InstanceNorm": normalization,
    "GroupNorm": normalization,
    # Torchlayers upsample
    "ConvPixelShuffle": upsample,
}


def __dir__():
    return [key for key in modules_map if not key.startswith("*")] + dir(torch.nn)


def __getattr__(name: str):
    def infer_module(module, tensor_index):
        klass = getattr(module, name, None)
        if klass is not None:
            return infer(klass, tensor_index)
        return None

    def process_entry(data) -> typing.Optional:
        if isinstance(data, typing.Iterable):
            if len(data) != 2:
                raise AttributeError(
                    f"Value of to-be-inferred module: {name} second argument has "
                    f"to be of length 2 but got {len(data)}. Check torchlayers documentation."
                )
            return infer_module(*data)

        if isinstance(data, types.ModuleType):
            return infer_module(data, 1)

        raise AttributeError(
            "Torchlayers recognized entry to infer but it's entry is of incorrect type. "
            "Check documentation about registration of user modules for more info."
        )

    def registered_module():
        for key, data in modules_map.items():
            if key.startswith("*"):
                klass = process_entry(data)
                if klass is not None:
                    return klass

        return None

    def noninferable_torchlayers():
        for module in (
            activations,
            convolution,
            normalization,
            pooling,
            regularization,
            preprocessing,
            upsample,
        ):
            klass = getattr(module, name, None)
            if klass is not None:
                return klass

        return None

    data = modules_map.get(name)
    if data is None:
        klass = registered_module()
    else:
        klass = process_entry(data)

    # Try to return from torchlayers without inference
    if klass is None:
        klass = noninferable_torchlayers()

    # Try to return from torch.nn without inference
    if klass is None:
        klass = getattr(torch.nn, name, None)

    # As far as we know there is no such module
    if klass is None:
        raise AttributeError(f"module {__name__} has no attribute {name}")
    return klass



import abc
import numbers
import random

import torch


class _GetInputs(torch.nn.Module):
    def _get_inputs(self, inputs):
        if self.inplace:
            return inputs
        return inputs.clone()


class RandomApply(_GetInputs):
    """Apply randomly a list of transformations with a given probability.


    .. note::
            **IMPORTANT**: This function is a `no-op` during inference phase
            (module in `eval` mode).

    Parameters
    ----------
    transforms : List | Tuple
        List of transformations
    p : float, optional
        Probability to apply list of transformations. Default: `0.5`
    inplace : bool, optional
        Whether to run this operation in-place. Default: `False`

    """

    def __init__(self, transforms, p: float = 0.5, inplace: bool = False):
        super().__init__()
        self.transforms = torch.nn.ModuleList(transforms)
        self.p: float = p
        self.inplace: bool = inplace

    def forward(self, inputs):
        if not self.training:
            return inputs

        x = self._get_inputs(inputs)
        if random.random() > self.p:
            return inputs
        for transform in self.transforms:
            x = transform(x)
        return x


class RandomChoice(torch.nn.Module):
    """Apply single transformation randomly picked from a list.

    .. note::
            **IMPORTANT**: This function is a `no-op` during inference phase
            (module in `eval` mode).

    Parameters
    ----------
    transforms : List | Tuple
        List of transformations
    """

    def __init__(self, transforms):
        super().__init__()
        self.transforms = torch.nn.ModuleList(transforms)

    def forward(self, inputs):
        if not self.training:
            return inputs

        transform = random.choice(self.transforms)
        return transform(inputs)


class RandomOrder(_GetInputs):
    """Apply single transformation randomly picked from a list.

    .. note::
            **IMPORTANT**: This function is a `no-op` during inference phase
            (module in `eval` mode).

    Parameters
    ----------
    transforms : List | Tuple
        List of transformations
    inplace : bool, optional
        Whether to run this operation in-place. Default: `False`
    """

    def __init__(self, transforms, inplace: bool = False):
        super().__init__()
        self.transforms = torch.nn.ModuleList(transforms)
        self.inplace: bool = inplace

    def forward(self, inputs):
        if not self.training:
            return inputs

        x = self._get_inputs(inputs)

        order = list(range(len(self.transforms)))
        random.shuffle(order)
        for i in order:
            x = self.transforms[i](x)
        return x


class Normalize(torch.nn.Module):
    """Normalize batch of tensor images with mean and standard deviation.

    Given mean values: `(M1,...,Mn)` and std values: `(S1,..,Sn)` for `n` channels
    (or other broadcastable to `n` values),
    this transform will normalize each channel of tensors in batch via formula:
    `output[channel] = (input[channel] - mean[channel]) / std[channel]`

    Parameters
    ----------
    mean : Tuple | List | torch.tensor
        Sequence of means for each channel
    std : Tuple | List | torch.tensor
        Sequence of means for each channel
    inplace : bool, optional
        Whether to run this operation in-place. Default: `False`

    """

    @classmethod
    def _transform_to_tensor(cls, tensor, name: str):
        if not torch.is_tensor(tensor):
            if isinstance(tensor, (tuple, list)):
                return torch.tensor(tensor)
            else:
                raise ValueError(
                    "{} is not an instance of either list, tuple or torch.tensor.".format(
                        name
                    )
                )
        return tensor

    @classmethod
    def _check_shape(cls, tensor, name):
        if len(tensor.shape) > 1:
            raise ValueError(
                "{} should be 0 or 1 dimensional tensor. Got {} dimensional tensor.".format(
                    name, len(tensor.shape)
                )
            )

    def __init__(self, mean: torch.Tensor, std: torch.Tensor, inplace: bool = False):
        tensor_mean = Normalize._transform_to_tensor(mean, "mean")
        tensor_std = Normalize._transform_to_tensor(std, "std")
        Normalize._check_shape(tensor_mean, "mean")
        Normalize._check_shape(tensor_std, "std")

        if torch.any(tensor_std == 0):
            raise ValueError(
                "One or more std values are zero which would lead to division by zero."
            )

        super().__init__()

        self.register_buffer("mean", tensor_mean)
        self.register_buffer("std", tensor_std)
        self.inplace: bool = inplace

    def forward(self, inputs):
        inputs_length = len(inputs.shape) - 2
        mean = self.mean.view(1, -1, *([1] * inputs_length))
        std = self.std.view(1, -1, *([1] * inputs_length))
        if self.inplace:
            inputs.sub_(mean).div_(std)
            return inputs
        return (inputs - mean) / std


class Transform(_GetInputs):
    """{header}

    {body}

    .. note::
            **IMPORTANT**: This function is a `no-op` during inference phase
            (module in `eval` mode).

    Parameters
    ----------
    p : float, optional
        Probability of applying transformation. Default: `0.5`
    batch : bool, optional
        Whether this operation should be applied on whole batch.
        If `True` the same transformation is applied on whole batch (or to
        no image in batch at all). If `False` apply transformation to `p` percent
        of images contained in batch at random. Default: `False`
    inplace : bool, optional
        Whether to run this operation in-place. Default: `False`

    """

    def __init__(self, p: float = 0.5, batch: bool = False, inplace: bool = False):
        if p < 0 or p > 1:
            raise ValueError("Probability of rotation should be between 0 and 1")
        super().__init__()
        self.p = p
        self.batch: bool = batch
        self.inplace: bool = inplace

    def forward(self, inputs):
        if self.training:
            x = self._get_inputs(inputs)
            if self.batch:
                if random.random() < self.p:
                    return self.transform(x)
            else:
                indices = torch.randperm(x.shape[0])[: int(x.shape[0] * self.p)]
                x[indices] = self.transform(x[indices])
                return x

        return inputs

    @abc.abstractmethod
    def transform(self, x):
        pass


class _RandomRotate90(Transform):
    """Randomly rotate image {} by `90` degrees `k` times.

    Rotation will be done in a clockwise manner.

    .. note::
            **IMPORTANT**: This function is a `no-op` during inference phase.

    Parameters
    ----------
    p : float, optional
        Probability of applying transformation. Default: `0.5`
    k : int, optional
        Number of times to rotate. Default: `1`
    batch : bool, optional
        Whether this operation should be applied on whole batch.
        If `True` the same transformation is applied on whole batch (or to
        no image in batch at all). If `False` apply transformation to `p` percent
        of images contained in batch at random. Default: `False`
    inplace : bool, optional
        Whether to run this operation in-place. Default: `False`

    """

    def __init__(
        self, p: float = 0.5, k: int = 1, batch: bool = False, inplace: bool = False
    ):
        super().__init__(p, batch, inplace)
        self.k: int = k


class ClockwiseRandomRotate90(_RandomRotate90):
    __doc__ = _RandomRotate90.__doc__.format("clockwise")

    def transform(self, x):
        return torch.rot90(x, k=self.k, dims=(-1, -2))


class AntiClockwiseRandomRotate90(_RandomRotate90):
    __doc__ = _RandomRotate90.__doc__.format("anticlockwise")

    def transform(self, x):
        return torch.rot90(x, k=self.k, dims=(-2, -1))


class RandomHorizontalFlip(Transform):
    __doc__ = Transform.__doc__.format(
        header="Randomly perform horizontal flip on batch of images.", body=""
    )

    def transform(self, x):
        return torch.flip(x, dims=(-1,))


class RandomVerticalFlip(Transform):
    __doc__ = Transform.__doc__.format(
        header="Randomly perform vertical flip on batch of images.", body=""
    )

    def transform(self, x):
        return torch.flip(x, dims=(-2,))


class RandomVerticalHorizontalFlip(Transform):
    __doc__ = Transform.__doc__.format(
        header="Randomly perform vertical and horizontal flip on batch of images.",
        body="",
    )

    def transform(self, x):
        return torch.flip(x, dims=(-2, -1))


class RandomErasing(Transform):
    """Randomly select rectangle regions in a batch of image and erase their pixels.

    Originally proposed by Zhong et al. in `Random Erasing Data Augmentation <https://arxiv.org/pdf/1708.04896.pdf>`__

    .. note::
            **IMPORTANT**: This function is a `no-op` during inference phase.
    .. note::
            Each image in batch will have the same rectangle region cut out
            due to efficiency reasons. It probably doesn't alter the idea
            drastically but exact effects weren't tested.


    Parameters
    ----------
    max_rectangles : int
        Maximum number of rectangles to create.
    max_height : int
        Maximum height of the rectangle.
    max_width : int, optional
        Maximum width of the rectangle. Default: same as `max_height`
    min_rectangles : int, optional
        Minimum number of rectangles to create. Default: same as `max_rectangles`
    min_height : int, optional
        Minimum height of the rectangle. Default: same as `max_height`
    min_width : int, optional
        Minimum width of the rectangle. Default: same as `min_width`
    fill : Callable, optional
        Callable used to fill the rectangle. It will be passed three arguments:
        `size` (as a `tuple`), `dtype`, `layout` and `device` of original tensor.
        If you want to specify `random uniform` filling you can use this:
        `lambda` function: `random = lambda size, dtype, layout, device: torch.randn(*size, dtype=dtype, layout=layout, device=device)`.
        If non-default is used, users are responsible for ensuring correct tensor format based
        on callable passed arguments.
        Default: fill with `0.0`.
    p : float, optional
        Probability of applying transformation. Default: `0.5`
    batch : bool, optional
        Whether this operation should be applied on whole batch.
        If `True` the same transformation is applied on whole batch (or to
        no image in batch at all). If `False` apply transformation to `p` percent
        of images contained in batch at random. Default: `False`
    inplace : bool, optional
        Whether to run this operation in-place. Default: `False`

    """

    @classmethod
    def _check_min_max(cls, minimum, maximum, name: str):
        if minimum > maximum:
            raise ValueError(
                "{} minimum is greater than maximum. Got minimum: {} and maximum: {}.".format(
                    name.capitalize(), minimum, maximum
                )
            )

    @classmethod
    def _check_greater_than_zero(cls, value, name: str):
        if value <= 0:
            raise ValueError(
                "Minimal {} should be greater than 0. Got {}".format(name, value)
            )

    @classmethod
    def _conditional_default(cls, value, default):
        if value is None:
            return default
        return value

    def __init__(
        self,
        max_rectangles: int,
        max_height: int,
        max_width: int = None,
        min_rectangles: int = None,
        min_height: int = None,
        min_width: int = None,
        fill=None,
        p: float = 0.5,
        batch: bool = False,
        inplace: bool = False,
    ):
        RandomErasing._check_greater_than_zero(min_rectangles, "holes")
        RandomErasing._check_greater_than_zero(min_height, "height")
        RandomErasing._check_greater_than_zero(min_width, "width")

        RandomErasing._check_min_max(min_rectangles, max_rectangles, "holes")
        RandomErasing._check_min_max(min_width, max_width, "width")
        RandomErasing._check_min_max(min_height, max_height, "height")

        self.max_rectangles: int = max_rectangles
        self.max_height: int = max_height

        self.max_width = RandomErasing._conditional_default(max_width, max_height)
        self.min_rectangles = RandomErasing._conditional_default(
            min_rectangles, max_rectangles
        )
        self.min_height = RandomErasing._conditional_default(min_height, max_height)
        self.min_width = RandomErasing._conditional_default(min_width, self.max_width)

        self.fill = RandomErasing._conditional_default(fill, lambda *_: 0.0)

        super().__init__(p, batch, inplace)

    def transform(self, x):
        holes = random.randint(self.min_rectangles, self.max_rectangles)

        start_hs = torch.randint(0, x.shape[-2] - self.max_height, (holes,))
        start_ws = torch.randint(0, x.shape[-1] - self.max_width, (holes,))
        heights = torch.randint(self.min_height, self.max_height, (holes,))
        widths = torch.randint(self.min_width, self.max_width, (holes,))

        for start_h, start_w, height, width in zip(start_hs, start_ws, heights, widths):
            x[..., start_h : start_h + height, start_w : start_w + width] = self.fill(
                (*(x.shape[:-2]), start_h + height, start_w + width),
                x.dtype,
                x.layout,
                x.device,
            )

        return x



import torch
import torchvision

import pytest
import torchfunc
import torchlayers as tl

# torchvision.transforms.RandomHorizontalFlip(p=1.0) vs tl.RandomHorizontalFlip
# torchvision.transforms.RandomVerticalFlip(p=1.0) vs tl.RandomVerticalFlip
# torchvision.transforms.Normalize vs tl.Normalize with same mean and std
# torchvision.transforms.functional.rotate via Lambdas with all Rotate90's
# RandomHorizontalVerticalFlip with both from torchvision one after another

# RandomApply withh the same probability and same outputs
# RandomChoice via same seed?
# RandomOrder via same seed?


class TorchvisionRotate:
    def __init__(self, angle):
        self.angle = angle

    def __call__(self, x):
        return torchvision.transforms.functional.to_tensor(
            torchvision.transforms.functional.rotate(x, self.angle)
        )


def generate_inputs():
    operations = [
        # Flipping
        (
            torchvision.transforms.Compose(
                [
                    torchvision.transforms.RandomHorizontalFlip(p=1.0),
                    torchvision.transforms.ToTensor(),
                ]
            ),
            tl.RandomHorizontalFlip(p=1.0),
        ),
        (
            torchvision.transforms.Compose(
                [
                    torchvision.transforms.RandomVerticalFlip(p=1.0),
                    torchvision.transforms.ToTensor(),
                ]
            ),
            tl.RandomVerticalFlip(p=1.0),
        ),
        (
            torchvision.transforms.Compose(
                [
                    torchvision.transforms.RandomVerticalFlip(p=1.0),
                    torchvision.transforms.RandomHorizontalFlip(p=1.0),
                    torchvision.transforms.ToTensor(),
                ]
            ),
            tl.RandomVerticalHorizontalFlip(p=1.0),
        ),
        # Normalization
        (
            torchvision.transforms.Compose(
                [
                    torchvision.transforms.ToTensor(),
                    torchvision.transforms.Normalize(
                        [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
                    ),
                ]
            ),
            tl.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
        ),
        # Random 90 rotations
        (TorchvisionRotate(90), tl.AntiClockwiseRandomRotate90(p=1.0),),
        (TorchvisionRotate(-90), tl.ClockwiseRandomRotate90(p=1.0),),
        (TorchvisionRotate(180), tl.ClockwiseRandomRotate90(p=1.0, k=2),),
    ]

    dataset = lambda transform: torchvision.datasets.CIFAR10(
        root=".", download=True, transform=transform
    )
    for torchvision_transform, torchlayers_module in operations:
        yield torch.utils.data.DataLoader(
            dataset(torchvision_transform), batch_size=2, shuffle=False
        ), torch.utils.data.DataLoader(
            dataset(torchvision.transforms.ToTensor()), batch_size=2, shuffle=False
        ), torchlayers_module


@pytest.mark.parametrize(
    "torchvision_dataloader,dataloader,module", list(generate_inputs())
)
def test_functionality(torchvision_dataloader, dataloader, module):
    for i, ((torchvision_images, _), (images, _)) in enumerate(
        zip(torchvision_dataloader, dataloader)
    ):
        if i == 10:
            break
        transformed_images = module(images)
        assert torch.allclose(transformed_images, torchvision_images)


# @pytest.mark.parametrize(
#     "torchvision_dataloader,dataloader,module",
#     list(generate_inputs(different_batches=True)),
# )
# def test_time(torchvision_dataloader, dataloader, module):
#     THRESHOLD = 1
#     LIMIT = 50

#     @torchfunc.Timer()
#     def time_torchvision(limit):
#         total = 0
#         for i, (torchvision_images, _) in enumerate(torchvision_dataloader):
#             if i == limit:
#                 break
#             total += torchvision_images.mean()
#         return total

#     @torchfunc.Timer()
#     def time_torchlayers(limit):
#         total = 0
#         for i, (images, _) in enumerate(dataloader):
#             if i == limit:
#                 break
#             total += module(images).mean()
#         return total

#     torchvision_value, torchvision_time = time_torchvision(LIMIT)
#     torchlayers_value, torchlayers_time = time_torchlayers(LIMIT)

#     assert torch.allclose(torchvision_value, torchlayers_value)
#     assert THRESHOLD * torchlayers_time < torchvision_time



import torch


def hard_sigmoid(tensor: torch.Tensor, inplace: bool = False) -> torch.Tensor:
    """
    Applies HardSigmoid function element-wise.

    See :class:`torchlayers.activations.HardSigmoid` for more details.

    Parameters
    ----------
    tensor : torch.Tensor
        Tensor activated element-wise
    inplace : bool, optional
        Whether operation should be performed `in-place`. Default: `False`

    Returns
    -------
    torch.Tensor
    """
    return torch.nn.functional.hardtanh(tensor, min_val=0, inplace=inplace)


class HardSigmoid(torch.nn.Module):
    """
    Applies HardSigmoid function element-wise.

    Uses `torch.nn.functional.hardtanh` internally with `0` and `1` ranges.

    Parameters
    ----------
    tensor : torch.Tensor
        Tensor activated element-wise

    """

    def forward(self, tensor: torch.Tensor):
        return hard_sigmoid(tensor)


def swish(tensor: torch.Tensor, beta: float = 1.0) -> torch.Tensor:
    """
    Applies Swish function element-wise.

    See :class:`torchlayers.activations.Swish` for more details.

    Parameters
    ----------
    tensor : torch.Tensor
        Tensor activated element-wise
    beta : float, optional
        Multiplier used for sigmoid. Default: 1.0 (no multiplier)

    Returns
    -------
    torch.Tensor
    """
    return torch.sigmoid(beta * tensor) * tensor


class Swish(torch.nn.Module):
    r"""
    Applies Swish function element-wise.

    .. math::

        Swish(x) = x / (1 + \exp(-beta * x))

    This form was originally proposed by Prajit Ramachandran et al. in
    `Searching for Activation Functions <https://arxiv.org/pdf/1710.05941.pdf>`__

    Parameters
    ----------
    beta : float, optional
        Multiplier used for sigmoid. Default: 1.0 (no multiplier)

    """

    def __init__(self, beta: float = 1.0):
        super().__init__()
        self.beta = beta

    def forward(self, tensor: torch.Tensor):
        return swish(tensor, self.beta)


def hard_swish(tensor: torch.Tensor) -> torch.Tensor:
    """
    Applies HardSwish function element-wise.

    See :class:`torchlayers.activations.HardSwish` for more details.

    Parameters
    ----------
    tensor : torch.Tensor
        Tensor activated element-wise

    Returns
    -------
    torch.Tensor
    """
    return tensor * torch.nn.functional.relu6(tensor + 3) / 6


class HardSwish(torch.nn.Module):
    r"""
    Applies HardSwish function element-wise.

    .. math::

        HardSwish(x) = x * \min(\max(0,x + 3), 6) / 6


    While similar in effect to `Swish` should be more CPU-efficient.
    Above formula proposed by in Andrew Howard et al. in `Searching for MobileNetV3 <https://arxiv.org/pdf/1905.02244.pdf>`__.

    Parameters
    ----------
    tensor : torch.Tensor
        Tensor activated element-wise

    """

    def forward(self, tensor: torch.Tensor):
        return hard_swish(tensor)



import pathlib

import setuptools


def read(HERE: pathlib.Path, filename, variable):
    namespace = {}

    exec(open(HERE / "torchlayers" / filename).read(), namespace)  # get version
    return namespace[variable]


HERE = pathlib.Path(__file__).resolve().parent

setuptools.setup(
    name=read(HERE, pathlib.Path("_name.py"), "_name"),
    version=read(HERE, pathlib.Path("_version.py"), "__version__"),
    license="MIT",
    author="Szymon Maszke",
    author_email="szymon.maszke@protonmail.com",
    description="Input shape inference and SOTA custom layers for PyTorch.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/torchlayers",
    packages=setuptools.find_packages(),
    install_requires=["torch>=1.3.0",],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    project_urls={
        "Website": "https://szymonmaszke.github.io/torchlayers",
        "Documentation": "https://szymonmaszke.github.io/torchlayers/#torchlayers",
        "Issues": "https://github.com/szymonmaszke/torchlayers/issues?q=is%3Aissue+is%3Aopen+sort%3Aupdated-desc",
    },
    keywords="pytorch keras input inference automatic shape layers sota custom imagenet resnet efficientnet",
)
